import asyncio
import aiohttp
import requests
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def http_request(url, payload=None):
    """ Makes a HTTP request to a website and returns its HTML content and full url address. """

    if payload is None:
        payload = {}

    content = requests.get(url, params=payload, verify=False)
    content.raise_for_status()  # Raise HTTPError for bad requests (4xx or 5xx)

    return content.text, content.url


class Connector(object):
    """ Used to make asynchronous HTTP requests. """

    def __init__(self, scrape_function, tasks):

        self.scrape_function = scrape_function
        self.tasks = tasks
        self.data = []

    async def __http_request__async(self, url, session):
        """ Sends asynchronous http request to URL address and scrapes the webpage. """

        async with session.get(url) as response:
            page_html = await response.read()

            return self.scrape_function(page_html, url)

    async def __async_scraper(self):
        """ Adds a URL's into a list of tasks and requests their response asynchronously. """

        async_tasks = []
        async with aiohttp.ClientSession() as session:
            for n in self.tasks:
                async_tasks.append(self.__http_request__async(n, session))

            self.data = await asyncio.gather(*async_tasks)

    def run_connector(self):
        """ Starts the asynchronous loop and returns the scraped data. """

        loop = asyncio.get_event_loop()
        loop.run_until_complete(self.__async_scraper())

        return self.data



from finviz.request_functions import Connector, http_request
from .save_data import export_to_db, export_to_csv
from urllib.parse import urlencode
from lxml import html
from lxml import etree
import finviz.scraper_functions as scrape

# TODO > Add __add__ and __slice__(?) methods to the Screener class


class Screener(object):
    """ Used to download data from http://www.finviz.com/screener.ashx. """

    def __init__(self, tickers=None, filters=None, rows=None, order='', signal='', table='Overview'):
        """
        Initilizes all variables to its values

        :param tickers: collection of ticker strings eg.: ['AAPL', 'AMD', 'WMT']
        :type tickers: list
        :param filters: collection of filters strings eg.: ['exch_nasd', 'idx_sp500', 'fa_div_none']
        :type filters: list
        :param rows: total number of rows to get
        :type rows: int
        :param order: table order eg.: '-price' (to sort table by descending price)
        :type order: str
        :param signal: show by signal eg.: 'n_majornews' (for stocks with major news)
        :type signal: str
        :param table: table type eg.: 'Performance'
        :type table: str
        :var self.data: pages containing data about each row inside a dictionary
        :type self.data: list
        """

        if tickers is None:
            self._tickers = []
        else:
            self._tickers = tickers

        if filters is None:
            self._filters = []
        else:
            self._filters = filters

        self._table_types = {
            'Overview': '110',
            'Valuation': '120',
            'Ownership': '130',
            'Performance': '140',
            'Custom': '150',
            'Financial': '160',
            'Technical': '170'
        }

        self._page_unparsed, self._url = http_request('https://finviz.com/screener.ashx', payload={
                                                   'v': self._table_types[table],
                                                   't': ','.join(self._tickers),
                                                   'f': ','.join(self._filters),
                                                   'o': order,
                                                   's': signal
                                                   })

        self._page_content = html.fromstring(self._page_unparsed)
        self._headers = self.__get_table_headers()

        if rows is None:
            self._rows = scrape.get_total_rows(self._page_content)
        else:
            self._rows = rows

        self.data = self.__search_screener()

    def __repr__(self):
        """ Returns a string containing readable representation of a table. """

        table_string = ''
        table_list = [self._headers]

        for page in self.data:
            for row in page:
                table_list.append([str(row[col] or '') for col in self._headers])

        col_size = [max(map(len, col)) for col in zip(*table_list)]
        format_str = ' | '.join(["{{:<{}}}".format(i) for i in col_size])
        table_list.insert(1, ['-' * i for i in col_size])

        for item in table_list:
            table_string += format_str.format(*item) + '\n'

        return table_string

    def __len__(self):
        """ Returns an int with the number of total rows. """

        return int(self._rows)

    def to_sqlite(self):
        """ Exports the generated table into a SQLite database, located in the user's current directory. """

        export_to_db(self._headers, self.data)

    def to_csv(self):
        """ Exports the generated table into a CSV file, located in the user's current directory. """

        export_to_csv(self._headers, self.data)

    def get_charts(self, period='d', size='l', chart_type='c', ta='1'):
        """
        Downloads the charts of tickers shown by the table.

        :param period: table period eg. : 'd', 'w' or 'm' for daily, weekly and monthly periods
        :type period: str
        :param size: table size eg.: 'l' for large or 's' for small - choose large for better quality but higher size
        :type size: str
        :param chart_type: chart type: 'c' for candles or 'l' for lines
        :type chart_type: str
        :param ta: technical analysis eg.: '1' to show ta '0' to hide ta
        :type ta: str
        """

        payload = {
            'ty': chart_type,
            'ta': ta,
            'p': period,
            's': size
        }

        base_url = 'https://finviz.com/chart.ashx?' + urlencode(payload)
        chart_urls = []

        for page in self.data:
            for row in page:
                chart_urls.append(base_url + '&t={}'.format(row.get('Ticker')))

        async_connector = Connector(scrape.download_chart_image, chart_urls)
        async_connector.run_connector()

    def __get_table_headers(self):
        """ Private function used to return the table headers. """

        first_row = self._page_content.cssselect('tr[valign="middle"]')

        headers = []
        for table_content in first_row[0]:

            if table_content.text is None:
                sorted_text_list = etree.tostring(table_content.cssselect('img')[0]).decode("utf-8").split('/>')
                headers.append(sorted_text_list[1])
            else:
                headers.append(table_content.text)

        return headers

    def __get_table_data(self, page=None, url=None):
        """ Private function used to return the table data from a single page. """

        def scrape_row(line):

            row_data = []

            for tags in line:
                if tags.text is not None:
                    row_data.append(tags.text)
                else:
                    row_data.append([span.text for span in tags.cssselect('span')][0])

            return row_data

        data_sets = []
        page = html.fromstring(page)
        all_rows = [i.cssselect('a') for i in page.cssselect('tr[valign="top"]')[1:]]

        for row in all_rows:

            if int(row[0].text) is self._rows:
                values = dict(zip(self._headers, scrape_row(row)))
                data_sets.append(values)
                break

            else:
                values = dict(zip(self._headers, scrape_row(row)))
                data_sets.append(values)

        return data_sets

    def __search_screener(self):
        """ Private function to return the data from the FinViz screener. """

        page_urls = scrape.get_page_urls(self._page_content, self._rows, self._url)
        async_connector = Connector(self.__get_table_data, page_urls)
        data = async_connector.run_connector()

        return data



import csv
import sqlite3
import re


def create_connection():
    """ Creates a database connection. """

    sqlite_file = "../screener.sqlite"

    try:
        conn = sqlite3.connect(sqlite_file)
        return conn
    except sqlite3.Error as error:
        raise ("An error has occurred while connecting to the database: ", error.args[0])


def export_to_csv(headers, data):
    """ Exports the generated table into a CSV file, located in the user's current directory. """

    with open('screener_results.csv', 'w', newline='') as output_file:
        dict_writer = csv.DictWriter(output_file, headers)
        dict_writer.writeheader()

        for n in data:
            dict_writer.writerows(n)


def export_to_db(headers, data):
    """ Exports the generated table into a SQLite database, located in the user's current directory. """

    field_list = ""
    table_name = "screener_results"  # name of the table to be created
    conn = create_connection()
    c = conn.cursor()

    for field in headers:

        field_cleaned = re.sub(r'[^\w\s]', '', field)
        field_cleaned = field_cleaned.replace(" ", "")
        field_list += field_cleaned + " TEXT, "

    c.execute("CREATE TABLE IF NOT EXISTS {tn} ({fl})"
              .format(tn=table_name, fl=field_list[:-2]))

    inserts = ""
    for data in data:

        for level in data:

            insert_fields = "("
            for field, value in level.items():
                insert_fields += "\"" + value + "\", "

            inserts += insert_fields[:-2] + "), "

    insert_lines = inserts[:-2]

    try:
        c.execute("INSERT INTO {tn} VALUES {iv}".
                  format(tn=table_name, iv=insert_lines))
    except sqlite3.Error as error:
        print("An error has occurred", error.args[0])

    conn.commit()
    conn.close()



from distutils.core import setup
setup(
  name = 'finviz',
  packages = ['finviz'],
  version = '1.1.2',
  license='MIT',
  description = 'Unofficial API for finviz.com',
  author = 'Mario Stoev',
  author_email = 'bg.mstoev@gmail.com',
  url = 'https://github.com/mariostoev/finviz',
  download_url = 'https://github.com/mariostoev/finviz/archive/1.1.2.tar.gz',
  keywords = ['finviz', 'api', 'screener', 'finviz api', 'charts', 'scraper'],
  install_requires=[
          'lxml',
          'requests',
          'aiohttp',
          'urllib3'
      ],
  classifiers=[
    'Development Status :: 5 - Production/Stable',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6',
  ],
)


from lxml import etree
import os


def get_total_rows(page_content):
    """ Returns the total number of rows(results). """

    total_element = page_content.cssselect('td[width="140"]')

    return int(etree.tostring(total_element[0]).decode("utf-8").split('</b>')[1].split(' ')[0])


def get_page_urls(page_content, rows, url):
    """ Returns a list containing all of the page URL addresses. """

    try:
        total_pages = int([i.text.split('/')[1] for i in page_content.cssselect('option[value="1"]')][0])
    except IndexError:
        raise Exception("No results matching the criteria: {}"
                        .format(url.split('?', 1)[1]))

    urls = []

    for page_number in range(1, total_pages + 1):

        sequence = 1 + (page_number - 1) * 20

        if sequence - 20 <= rows < sequence:
            break
        else:
            urls.append(url + '&r={}'.format(str(sequence)))

    return urls


def download_chart_image(page_content, url):
    """ Downloads a .jpg image of a chart into the "charts" folder. """

    file_name = url.split('t=')[1] + '.jpg'

    if not os.path.exists('charts'):
        os.mkdir('charts')

    with open('charts/' + file_name, 'wb') as handle:
        handle.write(page_content)
